use anyhow;

use std::{
    sync::mpsc::SyncSender,
    thread,
    time::Duration,
    io::Read,
};

use esp_idf_hal::prelude::*;
use esp_idf_hal::serial::{self, Rx, Tx};

use nmea0183::{Parser, ParseResult, Sentence, Source};

use crate::types::Msg;
use crate::serial::SerialIO;

struct GpsModule<UART: serial::Uart> {
    port: SerialIO<UART>,
}

impl<UART: serial::Uart> GpsModule<UART> {
    pub fn new(tx: Tx<UART>, rx: Rx<UART>) -> Self {
        GpsModule {
            port: SerialIO::new(tx, rx),
        }
    }
}

pub fn main<PRx,PTx>
(
    tx: PTx,
    rx: PRx,
    uart: serial::UART2,
    sender: SyncSender<Msg>,
) -> std::result::Result<(), anyhow::Error>
where
    PRx: esp_idf_hal::gpio::Pin + esp_idf_hal::gpio::InputPin + esp_idf_hal::gpio::OutputPin,
    PTx: esp_idf_hal::gpio::Pin + esp_idf_hal::gpio::InputPin + esp_idf_hal::gpio::OutputPin,
{
    let serial_pins = serial::Pins {
        tx,
        rx,
        cts: None,
        rts: None,
    };

    let serial: serial::Serial<serial::UART2, _, _> = serial::Serial::new(
        uart,
        serial_pins,
        serial::config::Config::default().baudrate(Hertz(9600)),
    )?;

    let (tx, rx) = serial.split();
    let mut device = GpsModule::new(tx, rx);

    let mut parser = Parser::new()
        .source_only(Source::GPS)
        .sentence_filter(Sentence::GLL | Sentence::GGA);

    let mut c = 0;
    let mut nmea = [0_u8; 1024];
    loop {
        if let Ok(_) = device.port.read(nmea.as_mut_slice()) {
            println!("\r\n\r\n\r\n\r\n");
            for result in parser.parse_from_bytes(&nmea[..]) {
                match result {
                    Ok(ParseResult::GLL(Some(gll))) => {
                        sender.send(Msg::Gps(gll.into()))?;
                    },
                    Ok(ParseResult::GGA(Some(gga))) => {
                        sender.send(Msg::Gps(gga.into()))?;
                    }
                    _ => { }
                }
            }
            c = 0;
        } else {
            println!("nothing to read after {} tries ...", c);
            if c > 100 {
                println!("reached {} retries ... bailing!", c);
                break;
            }
        }
        thread::sleep(Duration::from_millis(5000));
        c += 1;
    }

    Ok(())
}
